﻿
using System;
using System.Collections.Generic;
using System.Windows.Forms;


/* Complex root of a Polynomial according to Bairstow   */
/* x^5 - 9*x^4 + 33*x^3 + 63x^2 + 64x + 30 =            */
/* (x+1+j)*(x+1-j)*(x+2+j)*(x+2-j)*(x+3)                */
/* www.mosismath.com                                    */

namespace WindowsFormsApplication1
{

    public struct SRoot
    {
       public double p;
       public double q;
    }

    public partial class MainWin : Form
    {
        public const int order = 6; // number of reference points
        double[] a = new double[order];  // Polynomial to find the roots in
        public List<SRoot> root = new List<SRoot>();  // List of complex roots
        public SRoot r = new SRoot();

        public TextBox[] tbs;

        public MainWin()
        {
            int i;
            tbs = new TextBox[10];
          
            InitializeComponent();

            tbs[0] = textBoxY1; tbs[1] = textBoxY2; tbs[2] = textBoxY3; tbs[3] = textBoxY4; tbs[4] = textBoxY5;
            tbs[5] = textBoxY6; tbs[6] = textBoxY7; tbs[7] = textBoxY8; tbs[8] = textBoxY9; tbs[9] = textBoxY10;

            GMatrixA.RowCount = 2;

            // some sample values for a
            a[0] = 1.0;
            a[1] = 9.0;
            a[2] = 33.0;
            a[3] = 63.0;
            a[4] = 64.0;
            a[5] = 30.0;

            if (order < 10)
            {
                for (i = 0; i < order; i++)
                {
                    tbs[i].Text = Convert.ToString(a[i]);
                }
                for (i = order; i < 10; i++)
                {
                    tbs[i].Visible = false;
                }
            }
        }

        private int FindRoot(double[] a, int order)
        {
            double bn, bn1, bn2;
            double cn, cn1, cn2;
            double dxp, xp_old, dxq, xq_old;
            r = new SRoot();
            int i;
            int result = 1;
      
            r.p = -2.0; // Starting point
            r.q = -5.0;
            dxp = r.p;
            dxq = r.q;
            while ((Math.Abs(dxp) > 1E-10) && (Math.Abs(dxq) > 1E-10)) // check if we are close enough
            {
                xp_old = r.p;
                xq_old = r.q;
                bn2 = a[0];
                cn2 = bn2;
                bn1 = a[1] + r.p * bn2;
                cn1 = bn1 + r.p * cn2;
                cn = 0;
                bn = 0;
                for (i = 2; i < order; i++)
                {
                    bn = r.p * bn1 + r.q * bn2 + a[i];
                    if (i < order - 1)
                    {
                        cn = r.p * cn1 + r.q * cn2 + bn;
                        if (i < order - 2)
                        {
                            cn2 = cn1;
                            cn1 = cn;
                        }
                        bn2 = bn1;
                        bn1 = bn;
                    }
                }
                if (Math.Abs(cn1 * cn1 - cn * cn2) > 1E-30) // nex p and q
                {
                    r.p = r.p + (bn * cn2 - bn1 * cn1) / (cn1 * cn1 - cn * cn2);
                    r.q = r.q + (bn1 * cn - bn * cn1) / (cn1 * cn1 - cn * cn2);
                    dxp = r.p - xp_old;
                    dxq = r.q - xq_old;
                }
                else
                    result = -1;
            }
            if (result > 0)
                root.Add(r);
            return result;
        }

        private double[] Divide(double[] a, double[] b, int order)
        {
            int i,j;
            double fact;
            double[] a2 = new double[order - 1];
            for (i = 0; i < order - 3; i++)
            {
                if (Math.Abs(b[0]) > 1E-30)
                {
                    fact = a[i];
                    for (j = 0; j < 2; j++)
                    {
                        a2[j+i] = a[j + i + 1] - b[j + 1] * fact;
                    }
                    for (j = 2; j < order - 1-i; j++)
                    {
                        a2[j+i] = a[j + 1+i];
                    }

                    for (j = i; j < order - 1; j++)
                        a[j + 1] = a2[j];
                }
            }
            Array.Resize<double>(ref a, order - b.Length + 1);
            return a;
        }


        private void button1_Click(object sender, EventArgs e)
        {
            int iOrder = order;
            int i;
            root.Clear();
            double[] ax = new double[iOrder];
            double[] b = new double[3];
            double u, v;
            DataGridViewCell cell;

            if (order < 10)
            {
                for (i = 0; i < order; i++)
                {
                    a[i] = Convert.ToDouble(tbs[i].Text);
                }
            }
            for (i = 0; i < iOrder; i++)
                ax[i] = a[i];
            while (iOrder > 3)
            {
                if (FindRoot(ax, iOrder) >= 0)
                {
                    b[0] = 1;
                    b[1] = -r.p;
                    b[2] = -r.q;
                    ax = Divide(ax, b, iOrder);
                    iOrder = iOrder - 2;
                }
                else
                {
                    iOrder = 0;
                    MessageBox.Show("Calculation error");
                }
            }
            if (iOrder == 3)
            {
                r = new SRoot();
                r.p = ax[1] / ax[0];
                r.q = ax[2] / ax[0];
                root.Add(r);
                Array.Resize<double>(ref ax, 0);
            }

            if (root.Count > 0)
            {
                GMatrixA.RowCount = root.Count;
                for (i = 0; i < root.Count; i++)
                {
                    cell = GMatrixA[0, i];  // get cell to access
                    cell.Value = Convert.ToString(Convert.ToString(i + 1));
                    cell = GMatrixA[1, i];
                    cell.Value = Convert.ToString(Convert.ToString(Math.Round(root[i].p, 5)));
                    cell = GMatrixA[2, i];
                    cell.Value = Convert.ToString(Convert.ToString(Math.Round(root[i].q, 5)));
                    u = root[i].p / 2;
                    if ((-root[i].q - root[i].p * root[i].p / 4) > 0)
                    {
                        v = Math.Sqrt(-root[i].q - root[i].p * root[i].p / 4);
                        cell = GMatrixA[3, i];
                        cell.Value = Convert.ToString(Convert.ToString(Math.Round(u, 5)) + " + j" + Convert.ToString(Math.Round(v, 5)));
                        cell = GMatrixA[4, i];
                        cell.Value = Convert.ToString(Convert.ToString(Math.Round(u, 5)) + " - j" + Convert.ToString(Math.Round(v, 5)));
                    }
                    else
                    {
                        v = Math.Sqrt(root[i].p * root[i].p / 4 - root[i].q);
                        if (Math.Abs(-u + v) > 1e-10)
                        {
                            cell = GMatrixA[3, i];
                            cell.Value = Convert.ToString(Convert.ToString(Math.Round(-u + v, 5)));
                        }
                        if (Math.Abs(-u - v) > 1e-10)
                        {
                            cell = GMatrixA[4, i];
                            cell.Value = Convert.ToString(Convert.ToString(Math.Round(-u - v, 5)));
                        }
                    }
                }
                GRemainder.ColumnCount = ax.Length;
                for (i = 0; i < ax.Length; i++)
                {
                    GRemainder.Columns[i].HeaderText = "a" + Convert.ToString(i+1);
                    cell = GRemainder[i, 0];
                    cell.Value = Convert.ToString(Convert.ToString(Math.Round(ax[i], 5)));
                }
            }
            else
                MessageBox.Show("No complex roots :-(");
        }
    }
}
